%include "def.inc"

[bits 16]
org 0x7c00

  jmp short _start_boot
  nop

%include "fat32.inc"

_start_boot:

; init register
  mov ax, cs
  mov es, ax
  mov ds, ax
  mov ax, DEF_AREA_STACK_SEGMENT
  mov ss, ax
  mov sp, DEF_AREA_STACK_OFFSET
  xor ax, ax
  mov fs, ax

; clean up VAR's memory
  push es
  mov ax, DEF_AREA_VAR_SEGMENT
  mov es, ax
  mov di, DEF_AREA_VAR_OFFSET
  mov cx, DEF_AREA_VAR_END - DEF_AREA_VAR_START
  xor ax, ax
  cld
  rep stosb
  pop es

; initialize variable
  mov byte [fs:DEF_VAR_BOOT_DRIVE_NUMBER], dl
  mov byte [fs:DEF_VAR_DISPLAY_LINE], al
  mov word [fs:DEF_VAR_VBE_MODE], DEF_VBE_MODE
  mov byte [fs:DEF_VAR_DAP_SIZE], DEF_VAR_DAP_LENGTH

; clean screen
  mov ax, 0x0003
  int 10h

; msg: start boot
  mov bp, MSG_start_boot
  mov cx, MSG_start_boot_len
  call func_print

; read 4 KiB of FAT32 Root Entry
  mov word [fs:DEF_VAR_DAP_TOTAL], 0x08 ; 0x08 * 0x200 = 0x1000 bytes
  mov dword [fs:DEF_VAR_DAP_LBA_LOW], FAT32_ROOT_REGION
  mov word [fs:DEF_VAR_DAP_SEGMENT], DEF_AREA_FAT32_ROOT_ENTRY_SEGMENT
  mov word [fs:DEF_VAR_DAP_OFFSET], DEF_AREA_FAT32_ROOT_ENTRY_OFFSET
  call func_read_sector

; search loader.bin
  push es
  push word DEF_AREA_FAT32_ROOT_ENTRY_SEGMENT
  pop es
  mov edi, DEF_AREA_FAT32_ROOT_ENTRY_OFFSET
  mov esi, FILE_loader
  mov ecx, 11
  mov edx, 0x1000 >> 5 ; 32 bit per entries

.loop_search_loader:

  pushad ; protect ecx, edi, esi
  repe cmpsb
  popad
  je .loader_found

  add edi, 32
  dec edx
  jnz .loop_search_loader

  pop es
  mov bp, MSG_loader_not_found
  mov cx, MSG_loader_not_found_len
  call func_print
  jmp $

.loader_found:

  mov eax, dword [es:edi+0x1c]
  mov dword [fs:DEF_VAR_KERNEL_SIZE], eax
  push word [es:edi+0x14]
  push word [es:edi+0x1a]
  pop eax
  pop es

; read FAT32_SECTORS_PER_CLUSTER KiB of loader.bin
  sub eax, 2
  mov edx, FAT32_SECTORS_PER_CLUSTER
  mul edx
  add eax, FAT32_ROOT_REGION
  mov dword [fs:DEF_VAR_DAP_LBA_LOW], eax
  mov word [fs:DEF_VAR_DAP_TOTAL], FAT32_SECTORS_PER_CLUSTER ; 0x08 * 0x200 = 0x1000 bytes
  mov word [fs:DEF_VAR_DAP_SEGMENT], DEF_AREA_LOADER_SEGMENT
  mov word [fs:DEF_VAR_DAP_OFFSET], DEF_AREA_LOADER_OFFSET
  call func_read_sector

; integrity checking
  mov ax, DEF_AREA_LOADER_SEGMENT
  mov ds, ax
  cmp dword [ds:DEF_AREA_LOADER_OFFSET+0x03], 0x20000229
  jne .integrity_check_fail
  mov eax, dword [fs:DEF_VAR_KERNEL_SIZE]
  add eax, DEF_AREA_LOADER_OFFSET - 4
  cmp dword [ds:eax], 0x19991205
  jne .integrity_check_fail

; enter loader
  jmp DEF_AREA_LOADER_SEGMENT:DEF_AREA_LOADER_OFFSET

.integrity_check_fail:
  mov bp, MSG_loader_not_loaded
  mov cx, MSG_loader_not_loaded_len
  call func_print
  jmp $

;;;;;;; read sector from boot disk
;;;; !Required
;; [fs:DEF_VAR_DAP]
;; [fs:DEF_VAR_BOOT_DRIVE_NUMBER]
;;;;;;;
func_read_sector:
  push ds
  push fs
  pop ds
  mov esi, DEF_VAR_DAP
  mov ah, 42h
  mov dl, byte [fs:DEF_VAR_BOOT_DRIVE_NUMBER]
  int 13h
  pop ds
  jc .e_func_read_sector
  ret

.e_func_read_sector:
  mov bp, MSG_read_sector_e
  mov cx, MSG_read_sector_e_len
  call func_print
  jmp $


;;;;;;; print string
;; AL=01h: Assign all characters the attribute in BL; update cursor
;; BH=00h: Display page number
;; BL=0Fh: Attribute: font: light white; background: black; disable blinking
;; DL=00h: Display Column number
;;;; !Required
;; DH=[fs:DEF_VAR_DISPLAY_LINE]: Row
;;;; !Input
;; ES:BP: Points to string to be printed
;; CX: Number of characters in string
;;;; !Modified
;; [fs:DEF_VAR_DISPLAY_LINE]
;;;;;;;
func_print:
  mov ax, 0x1301
  mov bx, 0x000f
  mov dh, byte [fs:DEF_VAR_DISPLAY_LINE]
  xor dl, dl
  int 10h
  inc byte [fs:DEF_VAR_DISPLAY_LINE]
  ret


MSG_start_boot:     db   "start boot"
MSG_start_boot_len  equ  $ - MSG_start_boot

MSG_read_sector_e:     db   "read sector error"
MSG_read_sector_e_len  equ  $ - MSG_read_sector_e

MSG_loader_not_found:     db   "loader not found"
MSG_loader_not_found_len  equ  $ - MSG_loader_not_found

MSG_loader_not_loaded:     db   "loader not loaded"
MSG_loader_not_loaded_len  equ  $ - MSG_loader_not_loaded

FILE_loader:  db "LOADER  BIN"

  times 510-($-$$) db 0
  dw 0xAA55


