// @Author : Akvicor
// @Created Time : 2021-08-09 10:46:34
// @Description : generate fat32.inc

#include <stdio.h>
#include <stdlib.h>

unsigned char * buffer = NULL;

char * template =
";----------------------------------------------- [DEF]\n"
"\n"
"%%define FAT32_BYTES_PER_SECTOR     0x%08X\n"
"%%define FAT32_SECTORS_PER_CLUSTER  0x%08X\n"
"%%define FAT32_BYTES_PER_CLUSTER    0x%08X\n"
"%%define FAT32_FAT_REGION           0x%08X\n"
"%%define FAT32_ROOT_REGION          0x%08X\n"
"\n"
";----------------------------------------------- [OEM Manufacturer; offset: 0x03; length: 8 Bytes]\n"
"\n"
"BPB_OEMName               DB    \"AKVICOR \"              ; 8 Bytes\n"
"\n"
";----------------------------------------------- [BIOS Parameter Block; offset: 0x0b; length: 25 Bytes]\n"
"\n"
"; The number of bytes, in little-endian\n"
"BPB_BytesPerSec           DW    0x%04X                  ; offset: 0x0b[11]; length: 2  Bytes\n"
"; Number of Sectors per Cluster\n"
"BPB_SecPerClus            DB    0x%02X                    ; offset: 0x0d[13]; length: 1  Bytes\n"
"; Number of Reserved Sectors\n"
"BPB_RsvdSecCnt            DW    0x%04X                  ; offset: 0x0e[14]; length: 2  Bytes\n"
"; Number of File Allocation Tables\n"
"BPB_NumFATs               DB    0x%02X                    ; offset: 0x10[16]; length: 1  Bytes\n"
"; Root Entries. The total number file name\n"
"; entires than can be store in the root folder\n"
"; of the volume.\n"
"; For FAT32 volumes, this number is set to zero\n"
"BPB_RootEntCnt            DW    0x%04X                  ; offset: 0x11[17]; length: 2  Bytes\n"
"; 0 For FAT32\n"
"BPB_TotSec16              DW    0x%04X                  ; offset: 0x13[19]; length: 2  Bytes\n"
"; Media Type: F0 removable,\n"
"; F8 fixed media, i.e. hard disk\n"
"BPB_Media                 DB    0x%02X                    ; offset: 0x15[21]; length: 1  Bytes\n"
"; For FAT32, this field should be zero\n"
"BPB_FATSz16               DW    0x%04X                  ; offset: 0x16[22]; length: 2  Bytes\n"
"; Sectors per track\n"
"BPB_SecPerTrk             DW    0x%04X                  ; offset: 0x18[24]; length: 2  Bytes\n"
"; Number of heads\n"
"BPB_NumHeads              DW    0x%04X                  ; offset: 0x1A[26]; length: 2  Bytes\n"
"; Count of hidden sectors preceeding the\n"
"; partition that contains the FAT volume\n"
"BPB_HiddSec               DD    0x%08X              ; offset: 0x1C[28]; length: 4  Bytes\n"
"; Total count of sectors\n"
"BPB_TotSec32              DD    0x%08X              ; offset: 0x20[32]; length: 4  Bytes\n"
"\n"
";----------------------------------------------- [Extended BIOS Parameter Block; offset: 0x24; length: 54 Bytes]\n"
"\n"
"; 32-bit count of sectors occupied by one FAT\n"
"BPB_FATSz32               DD    0x%08X               ; offset: 0x24[36]; length: 4  Bytes\n"
"; Bits [0-3 ]: Zero based number of active FAT\n"
"; Bits [4-6 ]: Reserved\n"
"; Bits [7   ]:\n"
";  0: The FAT is mirrored at runtime into all FATs\n"
";  1: Only one FAT is active\n"
"; Bits [8-15]: Reserved\n"
"BPB_ExtFlags              DW    0x%04X                   ; offset: 0x28[40]; length: 2  Bytes\n"
"; High Byte: Major revision number\n"
"; The sample value prevents windows\n"
"; from loading this FAT partition\n"
"BPB_FSVer                 DW    0x%04X                   ; offset: 0x2A[42]; length: 2  Bytes\n"
"; Cluster number of the first cluster\n"
"; of the root directory\n"
"BPB_RootClus              DD    0x%08X               ; offset: 0x2C[44]; length: 4  Bytes\n"
"; File system info. Usually 1\n"
"BPB_FSInfo                DW    0x%04X                   ; offset: 0x30[48]; length: 2  Bytes\n"
"; If not zero, indicates the sector number\n"
"; in the reserved area of the volume of a\n"
"; copy of the boot record\n"
"BPB_BkBootSec             DW    0x%04X                   ; offset: 0x32[50]; length: 2  Bytes\n"
"; Reserved\n"
"BPB_Reserved              TIMES 12 DB 0                  ; offset: 0x34[52]; length: 12 Bytes\n"
"; Drive Number\n"
"BS_DrvNum                 DB    0x%02X                     ; offset: 0x40[64]; length: 1  Bytes\n"
"; Reserved\n"
"BS_Reserved1              DB    0x%02X                     ; offset: 0x41[65]; length: 1  Bytes\n"
"; Boot signature\n"
"BS_BootSig                DB    0x%02X                     ; offset: 0x42[66]; length: 1  Bytes\n"
"; Volume ID\n"
"BS_VolID                  DD    0x%08X               ; offset: 0x43[67]; length: 4  Bytes\n"
"; Volume Label\n"
"BS_VolLab                 DB    \"AOS    BOOT\"            ; offset: 0x47[71]; length: 11 Bytes\n"
"; Volume ID\n"
"BS_FilSysType             DB    \"FAT32   \"               ; offset: 0x52[82]; length: 8  Bytes\n"
"\n"
";----------------------------------------------- [Bootstrap Code; offset: 0x5A[90]; length: 420 Bytes]\n"
;

int read_file(char * filename, unsigned char * buffer, unsigned int size) {
  if(buffer == NULL) {
    printf("buffer is NULL!\n");
    return -1;
  }
  FILE *file = fopen(filename, "r");
  if(file == NULL) {
    printf("open file failed!\n");
    return -2;
  }
  fread(buffer, size, 1, file);
  fclose(file);
}

int convert_to_fat32(unsigned char * buffer) {
  if(buffer == NULL) {
    printf("buffer is NULL!\n");
    return -1;
  }

  // Bytes per Cluster
  unsigned int bytes_per_sector = *(unsigned short*)(buffer+0x0b);
  unsigned int sectors_per_cluster = *(unsigned char*)(buffer+0x0d);
  unsigned int bytes_per_cluster = bytes_per_sector * sectors_per_cluster;

  // FAT Region
  unsigned int reserved_sectors_num = *(unsigned short*)(buffer+0x0e);
  unsigned int fat_region = reserved_sectors_num;

  // Root Directory Region
  unsigned int fat_size = *(unsigned int*)(buffer+0x24);
  unsigned int fat_num = *(unsigned char*)(buffer+0x10);
  unsigned int total_fat_sectors = fat_size * fat_num;
  unsigned int root_directory_region = reserved_sectors_num + total_fat_sectors;

  printf(template,
      bytes_per_sector,
      sectors_per_cluster,
      bytes_per_cluster,
      fat_region,
      root_directory_region,
      *(unsigned short*)(buffer+0x0b),
      *(unsigned char*)(buffer+0x0d),
      *(unsigned short*)(buffer+0x0e),
      *(unsigned char*)(buffer+0x10),
      *(unsigned short*)(buffer+0x11),
      *(unsigned short*)(buffer+0x13),
      *(unsigned char*)(buffer+0x15),
      *(unsigned short*)(buffer+0x16),
      *(unsigned short*)(buffer+0x18),
      *(unsigned short*)(buffer+0x1a),
      *(unsigned int*)(buffer+0x1c),
      *(unsigned int*)(buffer+0x20),
      *(unsigned int*)(buffer+0x24),
      *(unsigned short*)(buffer+0x28),
      *(unsigned short*)(buffer+0x2a),
      *(unsigned int*)(buffer+0x2c),
      *(unsigned short*)(buffer+0x30),
      *(unsigned short*)(buffer+0x32),
      *(unsigned char*)(buffer+0x40),
      *(unsigned char*)(buffer+0x41),
      *(unsigned char*)(buffer+0x42),
      *(unsigned int*)(buffer+0x43)
    );
}

int main(int argc, char *argv[]){
  if (argc < 2) {
    printf("用法: %s <设备路径>\n", argv[0]);
    printf("例如: %s /dev/sdb\n", argv[0]);
    return 1;
  }

  buffer = (unsigned char *)calloc(512, sizeof(unsigned char));
  if (buffer == NULL) {
    printf("内存分配失败\n");
    return 1;
  }

  read_file(argv[1], buffer, 512);

  convert_to_fat32(buffer);

  free(buffer);
  return 0;
}

